# Copyright (C) 2021 Avery
#
# This file is part of py18n.
#
# py18n is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# py18n is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py18n.  If not, see <http://www.gnu.org/licenses/>.

from typing import List, Union

from .language import Language


class Py18nError(KeyError):
    pass


class InvalidLocaleError(Py18nError):
    def __init__(self, *args, locale: str) -> None:
        super().__init__(*args)
        self.locale = locale


class InvalidTranslationKeyError(Py18nError):
    def __init__(self, *args, key: str) -> None:
        super().__init__(*args)
        self.key = key


class I18n:
    def __init__(self, languages: List[Language], fallback: Union[str, int]) -> None:
        self._languages = {
            language.code: language
            for language in languages
        }

        self._fallback = None
        if isinstance(fallback, str):
            self._fallback = fallback

            if self._fallback not in self._languages:
                raise KeyError(
                    f"No language found with code {fallback} as fallback")
        elif isinstance(fallback, int):
            self._fallback = self._languages[languages[fallback].code]

        if self._fallback is None:
            raise KeyError(
                f"No fallback language set. Check documentation for correct usage")

    def get_text(
        self,
        key: str,
        locale: str,
        list_formatter: bool = None,
        use_translations: bool = True,
        should_fallback: bool = True,
        **kwargs
    ) -> str:
        """
        Wraps :func:`Language.get_text` to get translation based on the given locale

        .. seealso: documentation for :func:`Language.get_text`

        Parameters
        ----------
        key : str
            The key to search for
        list_formatter : bool, optional
            Function to format lists, by default None
        use_translations : bool, optional
            Whether to use translations in formatting, by default True
        should_fallback : bool, optional
            Should fallback to default locale, by default True

        Returns
        -------
        str
            Translated and formatted string

        Raises
        ------
        InvalidLocaleError
            If the locale does not exist on this instance
        InvalidTranslationKeyError
            If the key could not be found in the locale, nor in the fallback
            if `should_fallback` is `True`
        """
        # Get locale
        if locale not in self._languages:
            raise InvalidLocaleError(
                f"Given locale `{locale}` does not exist!", locale=locale)

        language = self._languages[locale]

        try:
            result = language.get_text(
                key, list_formatter=list_formatter, use_translations=use_translations, **kwargs)
        except KeyError as exc:
            # If we shouldn't fallback or this is the fallback
            if not should_fallback or locale == self._fallback:
                raise InvalidTranslationKeyError(
                    f"Translation {key} not found for {locale}!", key=key) from exc
        else:
            return result

        # We only get here if fallback is enabled and the text wasn't found in
        # the initial language
        try:
            result = self._languages[self._fallback].get_text(
                key, list_formatter=list_formatter, use_translations=use_translations, **kwargs)
        except KeyError as exc:
            raise InvalidTranslationKeyError(
                f"Translation {key} not found for {locale} nor fallback {self._fallback}", key=key) from exc
        else:
            return result



# Copyright (C) 2021 Avery
# 
# This file is part of py18n.
# 
# py18n is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# py18n is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with py18n.  If not, see <http://www.gnu.org/licenses/>.

from .i18n import I18n
from .language import Language
from .extension import I18nExtension

__version__ = "1.0.0"


# Copyright (C) 2021 Avery
#
# This file is part of py18n.
#
# py18n is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# py18n is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py18n.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict


class SafeDict(dict):
    def __missing__(self, key):
        return "{" + key + "}"


class Language:
    def __init__(self, name: str, code: str, translations: Dict[str, str]) -> None:
        self.name = name
        self.code = code
        self._translations = translations

    def _get_translation_from_key(self, key: str, raise_on_empty: bool = True) -> str:
        """
        Get the translation string from a given key. The default behaviour 
        supports simple key-translation access and dotted nesting.

        Parameters
        ----------
        key : str
            The key to parse
        raise_on_empty : bool, optional
            Whether to raise a KeyError when the returned value is an empty string, by default True

        Returns
        -------
        str
            The translation listed under the given key

        Raises
        ------
        KeyError
            The function attempted to access a key when using dotted nesting,
            but the value did not have this key
        KeyError
            If ``raise_on_empty`` is True, the value found is an empty string
        """
        if "." in key:
            parts = key.split(".")
            current = self._translations[parts[0]]
            for part in parts[1:]:
                if part in current:
                    current = current[part]
                else:
                    raise KeyError(f"{part} was not found under {current}")
        else:
            current = self._translations[key]

        if raise_on_empty and current == "":
            raise KeyError("Resultant string was empty")

        return current

    def join_list(self, value: list, connector: str) -> str:
        """
        Sensibly join list elements together

        Parameters
        ----------
        value : list
            The list of values to combine. Automatically converted to strings
        connector : str
            The text that goes either between two items when the list is 2
            items long, or between all but the last item of the list
            and the last item of the list when the list has more than 2
            items

        Returns
        -------
        str
            The list as a "sensible" string
        """
        # Stringify everything first
        value = [str(v) for v in value]
        if len(value) == 1:
            return value[0]
        elif len(value) == 2:
            # Just two items
            return connector.join(value)
        else:
            # All but last connected by ",", and last connected by connector
            return connector.join([','.join(value[:-1]), value[-1]])

    def and_(self, value: list, *args, **kwargs) -> str:
        """
        Wraps :func:`join_list` but uses the translation key ``and_``

        Parameters
        ----------
        value : list
            The list of values to combine. Automatically converted to strings

        Returns
        -------
        str
            The list as a "sensible" string
        """
        return self.join_list(value, " " + self._get_translation_from_key("and_",  *args, **kwargs) + " ")

    def or_(self, value: list, *args, **kwargs) -> str:
        """
        Wraps :func:`join_list` but uses the translation key ``or_``

        Parameters
        ----------
        value : list
            The list of values to combine. Automatically converted to strings

        Returns
        -------
        str
            The list as a "sensible" string
        """
        return self.join_list(value, " " + self._get_translation_from_key("or_",  *args, **kwargs) + " ")

    def get_text(
        self,
        key: str,
        list_formatter: bool = None,
        use_translations: bool = True,
        safedict=SafeDict,
        **kwargs
    ) -> str:
        """
        Get the formatted translation string

        Parameters
        ----------
        key : str
            The key to search for
        list_formatter : bool, optional
            Function to format lists, by default None

            .. seealso :: functions :func:`and_`, :func:`or_`, :func:`join_list`
        use_translations : bool, optional
            Whether to use translations in formatting, by default True

            For example, any missing parameters for the string (wrapped
            in curly braces) can be replaced by translations in the current
            language. This could be used to mix translation entries.

                >>> language = Language("English", "en", {
                    "you_lost": "You lost the {game}",
                    "game": "game",
                })

                >>> language.get_text("you_lost")
                "You lost the game"
        safedict : Any, optional
            Class to use as a "Safe dict", by default :cls:`SafeDict`
        **kwargs : dict, optional
            Parameters to pass to translation

        Returns
        -------
        str
            The translated string

        Raises
        ------
        KeyError
            The translation was not found (raised through `_get_translation_from_key`)
        """
        base_string = self._get_translation_from_key(key)

        # Sanitize passed arguments
        params = kwargs.copy()
        for key, value in params.items():
            if list_formatter and isinstance(value, list):
                params[key] = list_formatter(value)

        # Create the dict using given kwargs
        mapping = params
        if use_translations:
            # Put `**kwargs` after to prioritize given translations
            mapping = {
                **self._translations,
                **mapping
            }

        return base_string.format_map(safedict(**mapping))



# Copyright (C) 2021 Avery
#
# This file is part of py18n.
#
# py18n is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# py18n is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py18n.  If not, see <http://www.gnu.org/licenses/>.


from os import path
from setuptools import setup

with open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='Pycord18n',
      version='1.0.1',
      description='I18n for PyCord (A Discord.py fork)',
      author='starsflower, YoungTrep',
      url='https://github.com/YoungTrep/py18n',
      packages=['pycord18n'],
      package_dir={'pycord18n': './pycord18n'},
      include_package_data=True,
      install_requires=['py-cord'],
      extras_require={"voice": ["py-cord[voice]", "youtube-dl"]},

      # Description
      long_description=long_description,
      long_description_content_type='text/markdown'
)



# Copyright (C) 2021 Avery
#
# This file is part of py18n.
#
# py18n is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# py18n is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with py18n.  If not, see <http://www.gnu.org/licenses/>.


import contextvars
from typing import Callable, List, Optional, Union

from discord.ext import commands

from .i18n import I18n
from .language import Language


class I18nExtension(I18n):
    default_i18n_instance = None

    def __init__(
        self,
        languages: List[Language],
        fallback: Union[str, int],
        bot: Optional[commands.Bot] = None,
        get_locale_func: Callable = None,
        default: bool = True
    ) -> None:
        """
        Initialize the extension class.

        .. warning::

            The bot will only be attached to if both `bot` and `get_locale_func`
            are provided to this function. Otherwise it will not attach
            automatically.

        Parameters
        ----------
        languages : List[Language]
            List of lanugages to use
        fallback : Union[str, int]
            String ID or list index of the fallback locale
        bot : commands.Bot, optional
            The bot to attach to, by default None
        get_locale_func : Callable, optional
            If provided, init_bot will be run for you
        default : bool, optional
            Whether to make this i18n instance the default, by default True

            If there is no default i18n instance, this parameter is ignored and
            it is always set.

            The default is used by :func:`I18nExtension.contextual_get_text`.
        """
        super().__init__(languages, fallback)
        self._current_locale = contextvars.ContextVar("_current_locale")
        self._bot = None

        if default or I18nExtension.default_i18n_instance is None:
            I18nExtension.default_i18n_instance = self
        
        if self._bot and get_locale_func:
            self.init_bot(self._bot, get_locale_func)

    def init_bot(self, bot: commands.Bot, get_locale_func: Callable = None):
        """
        Initialize the given bot with the pre-invoke hooks to set the current
        context. 

        .. note ::

            Due to how discord.py works, this will override any previously
            set global pre-invoke hook.

            I recommend creating an override to have multiple pre- and post-
            invoke hooks if required, or setting the current locale yourself
            with :func:`set_current_locale`.

        Parameters
        ----------
        bot : commands.Bot
            The bot to attach to
        get_locale_func : Callable, optional
            The function that provides the locale code for the context, by default None

            It should take one argument, of type :cls:`discord.ext.commands.Context`
        """
        self._bot = bot
        if get_locale_func is None:
            # Just use the fallback
            get_locale_func = lambda *_: self._fallback

        async def pre(ctx):
            self.set_current_locale(get_locale_func(ctx))

        self._bot.before_invoke(pre)

    def set_current_locale(self, locale: str) -> str:
        """
        Set the current locale (for this context)

        Parameters
        ----------
        locale : str
            The locale
        """
        self._current_locale.set(locale)

    def get_current_locale(self) -> str:
        """
        Get the locale for this context, or the fallback locale if none is set

        Returns
        -------
        str
            The locale
        """
        return self._current_locale.get(self._fallback)

    @classmethod
    def contextual_get_text(
        cls,
        key: str,
        list_formatter: bool = None,
        use_translations: bool = True,
        should_fallback: bool = True,
        **kwargs
    ) -> str:
        """
        Wraps :func:`get_text` to use the current context's locale

        .. seealso: documentation for :func:`Language.get_text`

        Parameters
        ----------
        key : str
            The key to search for
        list_formatter : bool, optional
            Function to format lists, by default None
        use_translations : bool, optional
            Whether to use translations in formatting, by default True
        should_fallback : bool, optional
            Should fallback to default locale, by default True

        Returns
        -------
        str
            Translated and formatted string

        Raises
        ------
        NameError
            If there is no current i18n instance set
        """
        i18n = cls.default_i18n_instance
        if i18n is None:
            raise NameError("No default i18n instance has been initialized!")

        return i18n.get_text(
            key, i18n.get_current_locale(), list_formatter=list_formatter,
            use_translations=use_translations, should_fallback=should_fallback,
            **kwargs)


_ = I18nExtension.contextual_get_text
